# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class Ntuple(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsNtuple(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Ntuple()
        x.Init(buf, n + offset)
        return x

    # Ntuple
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Ntuple
    def Columns(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Column import Column
            obj = Column()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Ntuple
    def ColumnsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Ntuple
    def Instances(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .NtupleInstance import NtupleInstance
            obj = NtupleInstance()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Ntuple
    def InstancesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Ntuple
    def ColumnStatistics(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Statistics import Statistics
            obj = Statistics()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Ntuple
    def ColumnStatisticsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Ntuple
    def ColumnCovariances(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Covariance import Covariance
            obj = Covariance()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Ntuple
    def ColumnCovariancesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Ntuple
    def FunctionsLookup(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Ntuple
    def FunctionsLookupLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Ntuple
    def Functions(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .FunctionObject import FunctionObject
            obj = FunctionObject()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Ntuple
    def FunctionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def NtupleStart(builder): builder.StartObject(6)
def NtupleAddColumns(builder, columns): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(columns), 0)
def NtupleStartColumnsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def NtupleAddInstances(builder, instances): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(instances), 0)
def NtupleStartInstancesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def NtupleAddColumnStatistics(builder, columnStatistics): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(columnStatistics), 0)
def NtupleStartColumnStatisticsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def NtupleAddColumnCovariances(builder, columnCovariances): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(columnCovariances), 0)
def NtupleStartColumnCovariancesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def NtupleAddFunctionsLookup(builder, functionsLookup): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(functionsLookup), 0)
def NtupleStartFunctionsLookupVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def NtupleAddFunctions(builder, functions): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(functions), 0)
def NtupleStartFunctionsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def NtupleEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import unittest

import numpy

from aghast import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_loc_twodim(self):
        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(55, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()
        aloc = a.loc[7:, -3:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["7", "8", "9", "(-inf, 6]"]
        assert aloc.axis[1].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [[10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [20, 20, 20, 20, 20, 20, 40, 60]]
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()

        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(55, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()
        aloc = a.loc[..., -3:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert aloc.axis[1].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [[10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30]]
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()

        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(55, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()
        aloc = a.loc[:, -3:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert aloc.axis[1].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [[10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30]]
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()

        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(55, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()
        aloc = a.loc[5:9, -3:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert aloc.axis[1].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [[10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30], [10, 10, 10, 10, 10, 10, 20, 30]]
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()

        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(55, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()
        aloc = a.loc[7:]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["7", "8", "9", "(-inf, 6]"]
        assert aloc.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert aloc.counts.counts.array.tolist() == [[10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10], [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10], [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10], [20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20]]
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()

        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(55, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()
        aloc = a.loc[None, -3:2]
        assert len(aloc.axis) == 1
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [50, 50, 50, 50, 50, 50, 100, 150]
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()

        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(55, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()
        aloc = a.loc[7:, None]
        assert len(aloc.axis) == 1
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["7", "8", "9", "(-inf, 6]"]
        assert aloc.counts.counts.array.tolist() == [110, 110, 110, 220]
        assert a.axis[0].binning.toCategoryBinning().categories == ["5", "6", "7", "8", "9"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == (numpy.ones(55, dtype=int)*10).reshape((5, 11)).tolist()

    def test_loc_IntegerBinning(self):
        a = Histogram([Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(11, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]
        aloc = a.loc[-3:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 20, 30]
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]
        aloc = a.iloc[2:8]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 20, 30]
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]

        a = Histogram([Axis(IntegerBinning(-5, 5, loc_overflow=IntegerBinning.below1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([900, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[6, +inf)", "-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [900, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]
        aloc = a.loc[-3:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["-3", "-2", "-1", "0", "1", "2", "(-inf, -4]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 20, 930]
        assert a.axis[0].binning.toCategoryBinning().categories == ["[6, +inf)", "-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [900, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]

        a = Histogram([Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(11, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]
        aloc = a.loc[2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["2", "(-inf, 1]", "[3, +inf)"]
        assert aloc.counts.counts.array.tolist() == [10, 70, 30]
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]

        a = Histogram([Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.ones(11, dtype=int)*10)))
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]
        aloc = a.loc[None]
        assert len(aloc.axis) == 1
        assert aloc.axis[0].binning is None
        assert aloc.counts.counts.array.tolist() == [110]

    def test_loc_RegularBinning(self):
        a = Histogram([Axis(RegularBinning(20, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[-5, -4.5)", "[-4.5, -4)", "[-4, -3.5)", "[-3.5, -3)", "[-3, -2.5)", "[-2.5, -2)", "[-2, -1.5)", "[-1.5, -1)", "[-1, -0.5)", "[-0.5, 0)", "[0, 0.5)", "[0.5, 1)", "[1, 1.5)", "[1.5, 2)", "[2, 2.5)", "[2.5, 3)", "[3, 3.5)", "[3.5, 4)", "[4, 4.5)", "[4.5, 5)"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]

        aloc = a.iloc[::2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ['[-5, -4)', '[-4, -3)', '[-3, -2)', '[-2, -1)', '[-1, 0)', '[0, 1)', '[1, 2)', '[2, 3)', '[3, 4)', '[4, 5)']
        assert aloc.counts.counts.array.tolist() == [20, 20, 20, 20, 20, 20, 20, 20, 20, 20]
        aloc = a.iloc[0:10:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ['[-5, -4)', '[-4, -3)', '[-3, -2)', '[-2, -1)', '[-1, 0)', '[0, +inf)']
        assert aloc.counts.counts.array.tolist() == [20, 20, 20, 20, 20, 100]
        aloc = a.iloc[10:20:2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ['[0, 1)', '[1, 2)', '[2, 3)', '[3, 4)', '[4, 5)', '(-inf, 0)']
        assert aloc.counts.counts.array.tolist() == [20, 20, 20, 20, 20, 100]

        aloc = a.loc[-3.5::0.5]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[-3.5, -3)", "[-3, -2.5)", "[-2.5, -2)", "[-2, -1.5)", "[-1.5, -1)", "[-1, -0.5)", "[-0.5, 0)", "[0, 0.5)", "[0.5, 1)", "[1, 1.5)", "[1.5, 2)", "[2, 2.5)", "[2.5, 3)", "[3, 3.5)", "[3.5, 4)", "[4, 4.5)", "[4.5, 5)", "(-inf, -3.5)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 30]
        aloc = a.loc[::1]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ['[-5, -4)', '[-4, -3)', '[-3, -2)', '[-2, -1)', '[-1, 0)', '[0, 1)', '[1, 2)', '[2, 3)', '[3, 4)', '[4, 5)']
        assert aloc.counts.counts.array.tolist() == [20, 20, 20, 20, 20, 20, 20, 20, 20, 20]
        aloc = a.loc[-100:0:1]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ['[-5, -4)', '[-4, -3)', '[-3, -2)', '[-2, -1)', '[-1, 0)', '[0, +inf)']
        assert aloc.counts.counts.array.tolist() == [20, 20, 20, 20, 20, 100]
        aloc = a.loc[0:100:1]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ['[0, 1)', '[1, 2)', '[2, 3)', '[3, 4)', '[4, 5)', '(-inf, 0)']
        assert aloc.counts.counts.array.tolist() == [20, 20, 20, 20, 20, 100]

        a = Histogram([Axis(RegularBinning(19, RealInterval(-5, 4.5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[-5, -4.5)", "[-4.5, -4)", "[-4, -3.5)", "[-3.5, -3)", "[-3, -2.5)", "[-2.5, -2)", "[-2, -1.5)", "[-1.5, -1)", "[-1, -0.5)", "[-0.5, 0)", "[0, 0.5)", "[0.5, 1)", "[1, 1.5)", "[1.5, 2)", "[2, 2.5)", "[2.5, 3)", "[3, 3.5)", "[3.5, 4)", "[4, 4.5)"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]

        aloc = a.iloc[::2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ['[-5, -4)', '[-4, -3)', '[-3, -2)', '[-2, -1)', '[-1, 0)', '[0, 1)', '[1, 2)', '[2, 3)', '[3, 4)', '[4, +inf)']
        assert aloc.counts.counts.array.tolist() == [20, 20, 20, 20, 20, 20, 20, 20, 20, 10]

    def test_loc_EdgesBinning(self):
        a = Histogram([Axis(EdgesBinning([3.3, 5.5, 10.0, 55.5, 100.0]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([10, 10, 10, 10]))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[3.3, 5.5)", "[5.5, 10)", "[10, 55.5)", "[55.5, 100)"]
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10]

        aloc = a.iloc[::2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[3.3, 10)", "[10, 100)"]
        assert aloc.counts.counts.array.tolist() == [20, 20]

        aloc = a.loc[5.4:]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[3.3, 5.5)", "[5.5, 10)", "[10, 55.5)", "[55.5, 100)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10]

        aloc = a.loc[5.5:]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[5.5, 10)", "[10, 55.5)", "[55.5, 100)", "(-inf, 5.5)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10]

        aloc = a.loc[5.6:]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[5.5, 10)", "[10, 55.5)", "[55.5, 100)", "(-inf, 5.5)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10]

        aloc = a.loc[:55.6]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[3.3, 5.5)", "[5.5, 10)", "[10, 55.5)", "[55.5, 100)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10]

        aloc = a.loc[:55.5]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[3.3, 5.5)", "[5.5, 10)", "[10, 55.5)", "[55.5, +inf)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10]

        aloc = a.loc[:55.4]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[3.3, 5.5)", "[5.5, 10)", "[10, 55.5)", "[55.5, +inf)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 10]

        a = Histogram([Axis(EdgesBinning([3.3, 5.5, 10.0, 55.5]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([10, 10, 10]))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[3.3, 5.5)", "[5.5, 10)", "[10, 55.5)"]
        assert a.counts.counts.array.tolist() == [10, 10, 10]

        aloc = a.iloc[::2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[3.3, 10)", "[10, +inf)"]
        assert aloc.counts.counts.array.tolist() == [20, 10]

    def test_loc_IrregularBinning(self):
        a = Histogram([Axis(IrregularBinning([RealInterval(5, 10), RealInterval(8, 18), RealInterval(0, 1)]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([10, 10, 10]))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[5, 10)", "[8, 18)", "[0, 1)"]
        assert a.counts.counts.array.tolist() == [10, 10, 10]

        aloc = a.iloc[::2]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[5, 10)", "[0, 1)"]
        assert aloc.counts.counts.array.tolist() == [10, 10]

        aloc = a.loc[5:18]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[5, 10)", "[8, 18)", "(-inf, 5)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10]

        aloc = a.loc[6:17]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[5, 10)", "[8, 18)", "(-inf, 5)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10]

        aloc = a.loc[8:17]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[5, 10)", "[8, 18)", "(-inf, 5)"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10]

        aloc = a.loc[10:17]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[8, 18)", "(-inf, 8)"]
        assert aloc.counts.counts.array.tolist() == [10, 20]

    def test_loc_CategoryBinning(self):
        a = Histogram([Axis(CategoryBinning(["one", "two", "three", "four", "five"]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([10, 10, 10, 10, 10]))))
        assert a.axis[0].binning.loc_overflow == CategoryBinning.nonexistent
        assert a.counts.counts.array.tolist() == [10, 10, 10, 10, 10]

        aloc = a.iloc[::2]
        assert aloc.axis[0].binning.loc_overflow == CategoryBinning.above1
        assert aloc.axis[0].binning.categories == ["one", "three", "five"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 20]

        aloc = a.iloc[3:]
        assert aloc.axis[0].binning.loc_overflow == CategoryBinning.above1
        assert aloc.axis[0].binning.categories == ["four", "five"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 30]

        a = Histogram([Axis(CategoryBinning(["one", "two", "three", "four", "five"], loc_overflow=CategoryBinning.below1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([100, 10, 10, 10, 10, 10]))))
        assert a.axis[0].binning.loc_overflow == CategoryBinning.below1
        assert a.counts.counts.array.tolist() == [100, 10, 10, 10, 10, 10]

        aloc = a.iloc[::2]
        assert aloc.axis[0].binning.loc_overflow == CategoryBinning.above1
        assert aloc.axis[0].binning.categories == ["one", "three", "five"]
        assert aloc.counts.counts.array.tolist() == [10, 10, 10, 120]

        a = Histogram([Axis(CategoryBinning(["one", "two", "three", "four", "five"]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([1, 2, 3, 4, 5]))))
        assert a.axis[0].binning.loc_overflow == CategoryBinning.nonexistent
        assert a.counts.counts.array.tolist() == [1, 2, 3, 4, 5]

        aloc = a.loc[["four", "two", "three"]]
        assert aloc.axis[0].binning.loc_overflow == CategoryBinning.above1
        assert aloc.counts.counts.array.tolist() == [4, 2, 3, 6]

        a = Histogram([Axis(CategoryBinning(["one", "two", "three", "four", "five"], loc_overflow=CategoryBinning.below1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([100, 1, 2, 3, 4, 5]))))
        assert a.axis[0].binning.loc_overflow == CategoryBinning.below1
        assert a.counts.counts.array.tolist() == [100, 1, 2, 3, 4, 5]

        aloc = a.loc[["four", "two", "three"]]
        assert aloc.axis[0].binning.loc_overflow == CategoryBinning.above1
        assert aloc.counts.counts.array.tolist() == [4, 2, 3, 106]

    def test_loc_SparseRegularBinning(self):
        a = Histogram([Axis(SparseRegularBinning([5, 3, -2, 8, -5], 10.0))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([1, 2, 3, 4, 5]))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[50, 60)", "[30, 40)", "[-20, -10)", "[80, 90)", "[-50, -40)"]
        assert a.counts.counts.array.tolist() == [1, 2, 3, 4, 5]

        aloc = a.iloc[2:]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[-20, -10)", "[80, 90)", "[-50, -40)"]
        assert aloc.counts.counts.array.tolist() == [3, 4, 5]
        assert a.axis[0].binning.toCategoryBinning().categories == ["[50, 60)", "[30, 40)", "[-20, -10)", "[80, 90)", "[-50, -40)"]
        assert a.counts.counts.array.tolist() == [1, 2, 3, 4, 5]

        aloc = a.loc[0]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[0, 10)", "(-inf, 0)", "[0, +inf)"]
        assert aloc.counts.counts.array.tolist() == [0, 8, 7]

        aloc = a.loc[30:]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[50, 60)", "[30, 40)", "[80, 90)", "(-inf, 30)"]
        assert aloc.counts.counts.array.tolist() == [1, 2, 4, 8]
        assert a.axis[0].binning.toCategoryBinning().categories == ["[50, 60)", "[30, 40)", "[-20, -10)", "[80, 90)", "[-50, -40)"]
        assert a.counts.counts.array.tolist() == [1, 2, 3, 4, 5]

        aloc = a.loc[:30]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[-20, -10)", "[-50, -40)", "[30, +inf)"]
        assert aloc.counts.counts.array.tolist() == [3, 5, 7]
        assert a.axis[0].binning.toCategoryBinning().categories == ["[50, 60)", "[30, 40)", "[-20, -10)", "[80, 90)", "[-50, -40)"]
        assert a.counts.counts.array.tolist() == [1, 2, 3, 4, 5]

        aloc = a.loc[30::20]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[50, 70)", "[30, 50)", "[70, 90)", "(-inf, 30)"]
        assert aloc.counts.counts.array.tolist() == [1, 2, 4, 8]
        assert a.axis[0].binning.toCategoryBinning().categories == ["[50, 60)", "[30, 40)", "[-20, -10)", "[80, 90)", "[-50, -40)"]
        assert a.counts.counts.array.tolist() == [1, 2, 3, 4, 5]

        aloc = a.loc[::20]
        assert aloc.axis[0].binning.toCategoryBinning().categories == ["[40, 60)", "[20, 40)", "[-20, 0)", "[80, 100)", "[-60, -40)"]
        assert aloc.counts.counts.array.tolist() == [1, 2, 3, 4, 5]
        assert a.axis[0].binning.toCategoryBinning().categories == ["[50, 60)", "[30, 40)", "[-20, -10)", "[80, 90)", "[-50, -40)"]
        assert a.counts.counts.array.tolist() == [1, 2, 3, 4, 5]




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class EdgesBinning(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsEdgesBinning(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = EdgesBinning()
        x.Init(buf, n + offset)
        return x

    # EdgesBinning
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # EdgesBinning
    def Edges(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # EdgesBinning
    def EdgesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # EdgesBinning
    def EdgesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EdgesBinning
    def Overflow(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .RealOverflow import RealOverflow
            obj = RealOverflow()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # EdgesBinning
    def LowInclusive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return True

    # EdgesBinning
    def HighInclusive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # EdgesBinning
    def Circular(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def EdgesBinningStart(builder): builder.StartObject(5)
def EdgesBinningAddEdges(builder, edges): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(edges), 0)
def EdgesBinningStartEdgesVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def EdgesBinningAddOverflow(builder, overflow): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(overflow), 0)
def EdgesBinningAddLowInclusive(builder, lowInclusive): builder.PrependBoolSlot(2, lowInclusive, 1)
def EdgesBinningAddHighInclusive(builder, highInclusive): builder.PrependBoolSlot(3, highInclusive, 0)
def EdgesBinningAddCircular(builder, circular): builder.PrependBoolSlot(4, circular, 0)
def EdgesBinningEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class IntegerBinning(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsIntegerBinning(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = IntegerBinning()
        x.Init(buf, n + offset)
        return x

    # IntegerBinning
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # IntegerBinning
    def Min(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # IntegerBinning
    def Max(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # IntegerBinning
    def LocUnderflow(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # IntegerBinning
    def LocOverflow(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

def IntegerBinningStart(builder): builder.StartObject(4)
def IntegerBinningAddMin(builder, min): builder.PrependInt64Slot(0, min, 0)
def IntegerBinningAddMax(builder, max): builder.PrependInt64Slot(1, max, 0)
def IntegerBinningAddLocUnderflow(builder, locUnderflow): builder.PrependInt8Slot(2, locUnderflow, 0)
def IntegerBinningAddLocOverflow(builder, locOverflow): builder.PrependInt8Slot(3, locOverflow, 0)
def IntegerBinningEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import unittest

import pytest
import numpy as np

import aghast
from aghast import *

hist = pytest.importorskip("fnal_column_analysis_tools.hist")
pytest.importorskip("aghast._connect._fnalhist")

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_fnalhist(self):
        h_mascots = hist.Hist("fermi mascot showdown",
                              hist.Cat("animal", "type of animal"),
                              hist.Cat("vocalization", "onomatopoiea is that how you spell it?"),
                              hist.Bin("height", "height [m]", 10, 0, 5),
                              # weight is a reserved keyword
                              hist.Bin("mass", "weight (g=9.81m/s**2) [kg]", np.power(10., np.arange(5)-1)),
                             )
        adult_bison_h = np.random.normal(loc=2.5, scale=0.2, size=40)
        adult_bison_w = np.random.normal(loc=700, scale=100, size=40)
        h_mascots.fill(animal="bison", vocalization="huff", height=adult_bison_h, mass=adult_bison_w)
        goose_h = np.random.normal(loc=0.4, scale=0.05, size=1000)
        goose_w = np.random.normal(loc=7, scale=1, size=1000)
        h_mascots.fill(animal="goose", vocalization="honk", height=goose_h, mass=goose_w)
        crane_h = np.random.normal(loc=1, scale=0.05, size=4)
        crane_w = np.random.normal(loc=10, scale=1, size=4)
        h_mascots.fill(animal="crane", vocalization="none", height=crane_h, mass=crane_w)
        baby_bison_h = np.random.normal(loc=.5, scale=0.1, size=20)
        baby_bison_w = np.random.normal(loc=200, scale=10, size=20)
        baby_bison_cutefactor = 2.5*np.ones_like(baby_bison_w)
        h_mascots.fill(animal="bison", vocalization="baa", height=baby_bison_h, mass=baby_bison_w, weight=baby_bison_cutefactor)
        h_mascots.fill(animal="fox", vocalization="none", height=1., mass=30.)
        h_mascots.fill(animal="unicorn", vocalization="", height=np.nan, mass=np.nan)

        h_converted = aghast.to_fnalhist(aghast.from_fnalhist(h_mascots))
        for k in h_mascots.values().keys():
            assert k in h_converted.values().keys()
            assert np.all(h_mascots.values(overflow='allnan')[k] == h_converted.values(overflow='allnan')[k]), "mismatch for %r" % k



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class Statistics(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsStatistics(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Statistics()
        x.Init(buf, n + offset)
        return x

    # Statistics
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Statistics
    def Moments(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Moments import Moments
            obj = Moments()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Statistics
    def MomentsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Statistics
    def Quantiles(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Quantiles import Quantiles
            obj = Quantiles()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Statistics
    def QuantilesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Statistics
    def Mode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Modes import Modes
            obj = Modes()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Statistics
    def Min(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Extremes import Extremes
            obj = Extremes()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Statistics
    def Max(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Extremes import Extremes
            obj = Extremes()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def StatisticsStart(builder): builder.StartObject(5)
def StatisticsAddMoments(builder, moments): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(moments), 0)
def StatisticsStartMomentsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StatisticsAddQuantiles(builder, quantiles): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(quantiles), 0)
def StatisticsStartQuantilesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def StatisticsAddMode(builder, mode): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(mode), 0)
def StatisticsAddMin(builder, min): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(min), 0)
def StatisticsAddMax(builder, max): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(max), 0)
def StatisticsEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class PredicateBinning(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPredicateBinning(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PredicateBinning()
        x.Init(buf, n + offset)
        return x

    # PredicateBinning
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PredicateBinning
    def Predicates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # PredicateBinning
    def PredicatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # PredicateBinning
    def OverlappingFill(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

def PredicateBinningStart(builder): builder.StartObject(2)
def PredicateBinningAddPredicates(builder, predicates): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(predicates), 0)
def PredicateBinningStartPredicatesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def PredicateBinningAddOverlappingFill(builder, overlappingFill): builder.PrependInt8Slot(1, overlappingFill, 0)
def PredicateBinningEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import os.path

from setuptools import find_packages
from setuptools import setup

def get_version():
    g = {}
    exec(open(os.path.join("aghast", "version.py")).read(), g)
    return g["__version__"]

def get_description():
    return open("README.md").read()

setup(name = "aghast",
      version = get_version(),
      packages = find_packages(exclude = ["tests"]),
      scripts = [],
      data_files = [],
      description = "Aghast: aggregated, histogram-like statistics, sharable as Flatbuffers.",
      long_description = get_description(),
      long_description_content_type="text/markdown",
      author = "Jim Pivarski (IRIS-HEP)",
      author_email = "pivarski@princeton.edu",
      maintainer = "Jim Pivarski (IRIS-HEP)",
      maintainer_email = "pivarski@princeton.edu",
      url = "https://github.com/scikit-hep/aghast",
      download_url = "https://github.com/scikit-hep/aghast/releases",
      license = "BSD 3-clause",
      test_suite = "tests",
      install_requires = ["flatbuffers>=1.8.0", "numpy"],
      setup_requires = ["pytest-runner"],
      tests_require = ["pytest"],
      classifiers = [
          # "Development Status :: 1 - Planning",
          # "Development Status :: 2 - Pre-Alpha",
          # "Development Status :: 3 - Alpha",
          "Development Status :: 4 - Beta",
          # "Development Status :: 5 - Production/Stable",
          # "Development Status :: 6 - Mature",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ],
      platforms = "Any",
      )



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import numpy
try:
    import ROOT
except ImportError:
    raise ImportError("\n\nInstall ROOT package with:\n\n    conda install -c conda-forge root")

from aghast import *

def getbincontents(obj):
    if isinstance(obj, (ROOT.TH1C, ROOT.TH2C, ROOT.TH3C)):
        out = numpy.empty(obj.GetNcells(), dtype=numpy.int8)
        arraytype = "char"
    elif isinstance(obj, (ROOT.TH1S, ROOT.TH2S, ROOT.TH3S)):
        out = numpy.empty(obj.GetNcells(), dtype=numpy.int16)
        arraytype = "short"
    elif isinstance(obj, (ROOT.TH1I, ROOT.TH2I, ROOT.TH3I)):
        out = numpy.empty(obj.GetNcells(), dtype=numpy.int32)
        arraytype = "int"
    elif isinstance(obj, (ROOT.TH1F, ROOT.TH2F, ROOT.TH3F)):
        out = numpy.empty(obj.GetNcells(), dtype=numpy.float32)
        arraytype = "float"
    elif isinstance(obj, (ROOT.TH1D, ROOT.TH2D, ROOT.TH3D)):
        out = numpy.empty(obj.GetNcells(), dtype=numpy.float64)
        arraytype = "double"
    else:
        raise AssertionError(type(obj))

    name = "_getbincontents_{0}".format(type(obj).__name__)
    if name not in getbincontents.run:
        ROOT.gInterpreter.Declare("""
        void %s(%s* hist, %s* array) {
            int n = hist->GetNcells();
            for (int i = 0;  i < n;  i++) {
                array[i] = hist->GetBinContent(i);
            }
        }""" % (name, type(obj).__name__, arraytype))
        getbincontents.run[name] = getattr(ROOT, name)

    getbincontents.run[name](obj, out)
    return out

getbincontents.run = {}

def setbincontents(obj, array):
    if isinstance(obj, (ROOT.TH1C, ROOT.TH2C, ROOT.TH3C)):
        array = numpy.array(array, dtype=numpy.int8, copy=False)
        arraytype = "char"
    elif isinstance(obj, (ROOT.TH1S, ROOT.TH2S, ROOT.TH3S)):
        array = numpy.array(array, dtype=numpy.int16, copy=False)
        arraytype = "short"
    elif isinstance(obj, (ROOT.TH1I, ROOT.TH2I, ROOT.TH3I)):
        array = numpy.array(array, dtype=numpy.int32, copy=False)
        arraytype = "int"
    elif isinstance(obj, (ROOT.TH1F, ROOT.TH2F, ROOT.TH3F)):
        array = numpy.array(array, dtype=numpy.float32, copy=False)
        arraytype = "float"
    elif isinstance(obj, (ROOT.TH1D, ROOT.TH2D, ROOT.TH3D)):
        array = numpy.array(array, dtype=numpy.float64, copy=False)
        arraytype = "double"
    else:
        raise AssertionError(type(obj))

    assert array.shape == (obj.GetNcells(),)

    name = "_setbincontents_{0}".format(type(obj).__name__)
    if name not in setbincontents.run:
        ROOT.gInterpreter.Declare("""
        void %s(%s* hist, %s* array) {
            int n = hist->GetNcells();
            for (int i = 0;  i < n;  i++) {
                hist->SetBinContent(i, array[i]);
            }
        }""" % (name, type(obj).__name__, arraytype))
        setbincontents.run[name] = getattr(ROOT, name)

    setbincontents.run[name](obj, array)

setbincontents.run = {}

def to_root(obj, name):
    if isinstance(obj, Collection):
        raise NotImplementedError

    elif isinstance(obj, Histogram):
        axissummary = []
        slc = []
        for axis in obj.axis:
            if axis.binning is None:
                axissummary.append((RegularBinning, 1, 0.0, 1.0))
                slc.append(slice(None if axis.binning.overflow is None or axis.binning.overflow.loc_underflow == BinLocation.nonexistent else -numpy.inf,
                                 None if axis.binning.overflow is None or axis.binning.overflow.loc_overflow == BinLocation.nonexistent else numpy.inf))
            elif isinstance(axis.binning, IntegerBinning):
                axissummary.append((RegularBinning, 1 + axis.binning.max - axis.binning.min, axis.binning.min - 0.5, axis.binning.max + 0.5))
                slc.append(slice(None if axis.binning.overflow is None or axis.binning.overflow.loc_underflow == BinLocation.nonexistent else -numpy.inf,
                                 None if axis.binning.overflow is None or axis.binning.overflow.loc_overflow == BinLocation.nonexistent else numpy.inf))
            elif isinstance(axis.binning, RegularBinning):
                axissummary.append((RegularBinning, axis.binning.num, axis.binning.interval.low, axis.binning.interval.high))
                slc.append(slice(None if axis.binning.overflow is None or axis.binning.overflow.loc_underflow == BinLocation.nonexistent else -numpy.inf,
                                 None if axis.binning.overflow is None or axis.binning.overflow.loc_overflow == BinLocation.nonexistent else numpy.inf))
            elif isinstance(axis.binning, HexagonalBinning):
                raise TypeError("no ROOT equivalent for HexagonalBinning")
            elif isinstance(axis.binning, EdgesBinning):
                axissummary.append((EdgesBinning, numpy.array(axis.binning.edges, dtype=numpy.float64, copy=False)))
                slc.append(slice(None if axis.binning.overflow is None or axis.binning.overflow.loc_underflow == BinLocation.nonexistent else -numpy.inf,
                                 None if axis.binning.overflow is None or axis.binning.overflow.loc_overflow == BinLocation.nonexistent else numpy.inf))
            elif isinstance(axis.binning, IrregularBinning):
                axissummary.append((CategoryBinning, axis.binning.toCategoryBinning().categories))
                slc.append(slice(None))
            elif isinstance(axis.binning, CategoryBinning):
                axissummary.append((CategoryBinning, axis.binning.categories))
                slc.append(slice(None))
            elif isinstance(axis.binning, SparseRegularBinning):
                axissummary.append((CategoryBinning, axis.binning.toCategoryBinning().categories))
                slc.append(slice(None))
            elif isinstance(axis.binning, FractionBinning):
                axissummary.append((CategoryBinning, axis.binning.toCategoryBinning().categories))
                slc.append(slice(None))
            elif isinstance(axis.binning, PredicateBinning):
                axissummary.append((CategoryBinning, axis.binning.toCategoryBinning().categories))
                slc.append(slice(None))
            elif isinstance(axis.binning, VariationBinning):
                axissummary.append((CategoryBinning, axis.binning.toCategoryBinning().categories))
                slc.append(slice(None))
            else:
                raise AssertionError(type(axis.binning))
                
        if isinstance(obj.axis[0].binning, FractionBinning):
            raise NotImplementedError

        sumw = obj.counts[tuple(slc)]
        if isinstance(sumw, dict):
            sumw, sumw2 = sumw["sumw"], sumw["sumw2"]
            sumw2 = numpy.array(sumw2, dtype=numpy.float64, copy=False)
        else:
            sumw2 = None

        if len(obj.profile) != 0:
            raise NotImplementedError

        if issubclass(sumw.dtype.type, numpy.int8):
            cls = ROOT.TH1C
            sumw = numpy.array(sumw, dtype=numpy.int8, copy=False)
        elif issubclass(sumw.dtype.type, (numpy.uint8, numpy.int16)):
            cls = ROOT.TH1S
            sumw = numpy.array(sumw, dtype=numpy.int16, copy=False)
        elif issubclass(sumw.dtype.type, (numpy.uint16, numpy.int32)):
            cls = ROOT.TH1I
            sumw = numpy.array(sumw, dtype=numpy.int32, copy=False)
        elif issubclass(sumw.dtype.type, numpy.float32):
            cls = ROOT.TH1F
            sumw = numpy.array(sumw, dtype=numpy.float32, copy=False)
        else:
            cls = ROOT.TH1D
            sumw = numpy.array(sumw, dtype=numpy.float64, copy=False)

        title = "" if obj.title is None else obj.title

        if len(axissummary) == 1:
            if axissummary[0][0] is RegularBinning:
                num, low, high = axissummary[0][1:]
                out = cls(name, title, num, low, high)
                xaxis = out.GetXaxis()
                
            elif axissummary[0][0] is EdgesBinning:
                edges = axissummary[0][1]
                out = cls(name, title, len(edges) - 1, edges)
                xaxis = out.GetXaxis()

            elif axissummary[0][0] is CategoryBinning:
                categories = axissummary[0][1]
                out = cls(name, title, len(categories), 0.5, len(categories) + 0.5)

            else:
                raise AssertionError(axissummary[0][0])

            assert len(sumw.shape) == 1
            if sumw2 is not None:
                assert sumw.shape == sumw2.shape

            if len(sumw) == out.GetNbinsX():
                sumw = numpy.concatenate([numpy.array([0], dtype=sumw.dtype), sumw, numpy.array([0], dtype=sumw.dtype)])
                if sumw2 is not None:
                    sumw2 = numpy.concatenate([numpy.array([0], dtype=sumw2.dtype), sumw2, numpy.array([0], dtype=sumw2.dtype)])
            elif len(sumw) == out.GetNbinsX() + 1 and obj.axis[0].loc_underflow == BinLocation.nonexistent:
                sumw = numpy.concatenate([numpy.array([0], dtype=sumw.dtype), sumw])
                if sumw2 is not None:
                    sumw2 = numpy.concatenate([numpy.array([0], dtype=sumw2.dtype), sumw2])
            elif len(sumw) == out.GetNbinsX() + 1:
                sumw = numpy.concatenate([sumw, numpy.array([0], dtype=sumw.dtype)])
                if sumw2 is not None:
                    sumw2 = numpy.concatenate([sumw2, numpy.array([0], dtype=sumw2.dtype)])
            elif len(sumw) != out.GetNbinsX() + 2:
                raise AssertionError((len(sumw), out.GetNbinsX() + 2))

            setbincontents(out, sumw)

            if sumw2 is not None:
                out.Sumw2()
                sumw2obj = out.GetSumw2()
                setbincontents(sumw2obj, sumw2)

            if axissummary[0][0] is CategoryBinning:
                xaxis = out.GetXaxis()
                for i, x in enumerate(categories):
                    xaxis.SetBinLabel(i + 1, x)

            numentries = None
            stats0 = None
            stats1 = None
            stats2 = None
            stats3 = None
            if len(obj.axis[0].statistics) != 0:
                for moment in obj.axis[0].statistics[0].moments:
                    sumwxn = moment.sumwxn.flatarray
                    if moment.n == 0 and moment.weightpower == 0 and len(sumwxn) == 1:
                        numentries, = sumwxn
                    if moment.n == 0 and moment.weightpower == 1 and len(sumwxn) == 1:
                        stats0, = sumwxn
                    if moment.n == 0 and moment.weightpower == 2 and len(sumwxn) == 1:
                        stats1, = sumwxn
                    if moment.n == 1 and moment.weightpower == 1 and len(sumwxn) == 1:
                        stats2, = sumwxn
                    if moment.n == 2 and moment.weightpower == 1 and len(sumwxn) == 1:
                        stats3, = sumwxn

            if numentries is None:
                if isinstance(obj.counts, UnweightedCounts):
                    numentries = obj.counts.flatarray.sum()
                elif obj.counts.unweighted is not None:
                    numentries = obj.counts.unweighted.flatarray.sum()
                else:
                    numentries = obj.counts.sumw.flatarray.sum()

            out.SetEntries(numentries)

            if stats0 is not None and stats1 is not None and stats2 is not None and stats3 is not None:
                stats = numpy.array([stats0, stats1, stats2, stats3], dtype=numpy.float64)
                out.PutStats(stats)

            xaxis.SetTitle("" if obj.axis[0].title is None else obj.axis[0].title)

            return out

        elif len(axissummary) == 2:
            raise NotImplementedError

        elif len(axissummary) == 3:
            raise NotImplementedError

        else:
            raise NotImplementedError

    elif isinstance(obj, Ntuple):
        raise NotImplementedError

    else:
        raise TypeError("cannot convert {0}".format(type(obj)))

def from_root(obj, collection=False):
    if isinstance(obj, ROOT.TH1):
        if isinstance(obj, ROOT.TProfile):
            raise NotImplementedError

        elif isinstance(obj, ROOT.TProfile2D):
            raise NotImplementedError

        elif isinstance(obj, ROOT.TProfile3D):
            raise NotImplementedError

        if not isinstance(obj, (ROOT.TH2, ROOT.TH3)):
            sumwarray = getbincontents(obj)

            xaxis = obj.GetXaxis()
            labels = xaxis.GetLabels()
            if labels:
                categories = [str(x) for x in labels]
            else:
                categories = None

            if obj.GetSumw2N() != 0:
                sumw2obj = h.GetSumw2()
                sumw2array = getbincontents(sumw2obj)
                if categories is not None:
                    sumwarray = sumwarray[1 : len(categories) + 1]
                    sumw2array = sumw2array[1 : len(categories) + 1]
                counts = WeightedCounts(
                    sumw=InterpretedInlineBuffer.fromarray(sumwarray),
                    sumw2=InterpretedInlineBuffer.fromarray(sumw2array))
            else:
                if categories is not None:
                    sumwarray = sumwarray[1 : len(categories) + 1]
                counts = UnweightedCounts(
                    counts=InterpretedInlineBuffer.fromarray(sumwarray))

            if categories is not None:
                binning = CategoryBinning(categories)

            elif xaxis.IsVariableBinSize():
                num = obj.GetNbinsX()
                edges = numpy.empty(num + 1, dtype=numpy.float64)
                xaxis.GetLowEdge(edges)
                edges[-1] = xaxis.GetBinUpEdge(num)
                binning = EdgesBinning(edges, overflow=RealOverflow(loc_underflow=BinLocation.below1, loc_overflow=BinLocation.above1))

            else:
                num = obj.GetNbinsX()
                low = xaxis.GetBinLowEdge(1)
                high = xaxis.GetBinUpEdge(num)
                binning = RegularBinning(num, RealInterval(low, high), overflow=RealOverflow(loc_underflow=BinLocation.below1, loc_overflow=BinLocation.above1))

            stats = numpy.zeros(4, numpy.float64)
            obj.GetStats(stats)
            entries = Moments(InterpretedInlineBuffer.fromarray(numpy.array([obj.GetEntries()], dtype=numpy.int64)), n=0, weightpower=0)
            sumw = Moments(InterpretedInlineBuffer.fromarray(stats[0:1]), n=0, weightpower=1)
            sumw2 = Moments(InterpretedInlineBuffer.fromarray(stats[1:2]), n=0, weightpower=2)
            sumwx = Moments(InterpretedInlineBuffer.fromarray(stats[2:3]), n=1, weightpower=1)
            sumwx2 = Moments(InterpretedInlineBuffer.fromarray(stats[3:4]), n=2, weightpower=1)
            statistics = [Statistics(moments=[entries, sumw, sumw2, sumwx, sumwx2])]

            title = xaxis.GetTitle()
            if title == "":
                title = None
            axis = Axis(binning, statistics=statistics, title=title)

            title = obj.GetTitle()
            if title == "":
                title = None
            out = Histogram([axis], counts, title=title)

            if collection:
                raise NotImplementedError
            else:
                return out

        elif isinstance(obj, ROOT.TH2):
            raise NotImplementedError

        elif isinstance(obj, ROOT.TH3):
            raise NotImplementedError

        else:
            raise TypeError("cannot convert {0}".format(type(obj).__name__))

    elif isinstance(obj, ROOT.TEfficiency):
        raise NotImplementedError

    elif isinstance(obj, ROOT.TTree):
        raise NotImplementedError

    else:
        raise TypeError("cannot convert {0}".format(type(obj).__name__))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class InterpretedBuffer(object):
    NONE = 0
    InterpretedInlineBuffer = 1
    InterpretedInlineInt64Buffer = 2
    InterpretedInlineFloat64Buffer = 3
    InterpretedExternalBuffer = 4




#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import unittest

import numpy

import aghast
from aghast import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_numpy1d(self):
        before = numpy.histogram(numpy.random.normal(0, 1, int(1e6)), bins=100, range=(-5, 5))
        after = aghast.to_numpy(aghast.from_numpy(before))
        assert numpy.array_equal(before[0], after[0])
        assert numpy.array_equal(before[1], after[1])

    def test_numpy2d(self):
        before = numpy.histogram2d(x=numpy.random.normal(0, 1, int(1e6)), y=numpy.random.normal(0, 1, int(1e6)), bins=(10, 10), range=((-5, 5), (-5, 5)))
        after = aghast.to_numpy(aghast.from_numpy(before))
        assert numpy.array_equal(before[0], after[0])
        assert numpy.array_equal(before[1], after[1])
        assert numpy.array_equal(before[2], after[2])

    def test_numpydd(self):
        before = numpy.histogramdd((numpy.random.normal(0, 1, int(1e6)), numpy.random.normal(0, 1, int(1e6)), numpy.random.normal(0, 1, int(1e6))), bins=(5, 5, 5), range=((-5, 5), (-5, 5), (-5, 5)))
        after = aghast.to_numpy(aghast.from_numpy(before))
        assert numpy.array_equal(before[0], after[0])
        assert numpy.array_equal(before[1][0], after[1][0])
        assert numpy.array_equal(before[1][1], after[1][1])
        assert numpy.array_equal(before[1][2], after[1][2])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class Filter(object):
    filter_none = 0
    filter_gzip = 1
    filter_lzma = 2
    filter_lz4 = 3




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class HexagonalBinning(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsHexagonalBinning(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = HexagonalBinning()
        x.Init(buf, n + offset)
        return x

    # HexagonalBinning
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # HexagonalBinning
    def Qmin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # HexagonalBinning
    def Qmax(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # HexagonalBinning
    def Rmin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # HexagonalBinning
    def Rmax(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # HexagonalBinning
    def Coordinates(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # HexagonalBinning
    def Xorigin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # HexagonalBinning
    def Yorigin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # HexagonalBinning
    def Qangle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # HexagonalBinning
    def BinWidth(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 1.0

    # HexagonalBinning
    def Qoverflow(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = o + self._tab.Pos
            from .RealOverflow import RealOverflow
            obj = RealOverflow()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # HexagonalBinning
    def Roverflow(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            x = o + self._tab.Pos
            from .RealOverflow import RealOverflow
            obj = RealOverflow()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def HexagonalBinningStart(builder): builder.StartObject(11)
def HexagonalBinningAddQmin(builder, qmin): builder.PrependInt64Slot(0, qmin, 0)
def HexagonalBinningAddQmax(builder, qmax): builder.PrependInt64Slot(1, qmax, 0)
def HexagonalBinningAddRmin(builder, rmin): builder.PrependInt64Slot(2, rmin, 0)
def HexagonalBinningAddRmax(builder, rmax): builder.PrependInt64Slot(3, rmax, 0)
def HexagonalBinningAddCoordinates(builder, coordinates): builder.PrependInt8Slot(4, coordinates, 0)
def HexagonalBinningAddXorigin(builder, xorigin): builder.PrependFloat64Slot(5, xorigin, 0.0)
def HexagonalBinningAddYorigin(builder, yorigin): builder.PrependFloat64Slot(6, yorigin, 0.0)
def HexagonalBinningAddQangle(builder, qangle): builder.PrependFloat64Slot(7, qangle, 0.0)
def HexagonalBinningAddBinWidth(builder, binWidth): builder.PrependFloat64Slot(8, binWidth, 1.0)
def HexagonalBinningAddQoverflow(builder, qoverflow): builder.PrependStructSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(qoverflow), 0)
def HexagonalBinningAddRoverflow(builder, roverflow): builder.PrependStructSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(roverflow), 0)
def HexagonalBinningEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class CategoryBinning(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsCategoryBinning(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = CategoryBinning()
        x.Init(buf, n + offset)
        return x

    # CategoryBinning
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CategoryBinning
    def Categories(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # CategoryBinning
    def CategoriesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # CategoryBinning
    def LocOverflow(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

def CategoryBinningStart(builder): builder.StartObject(2)
def CategoryBinningAddCategories(builder, categories): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(categories), 0)
def CategoryBinningStartCategoriesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def CategoryBinningAddLocOverflow(builder, locOverflow): builder.PrependInt8Slot(1, locOverflow, 0)
def CategoryBinningEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class SparseRegularBinning(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSparseRegularBinning(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SparseRegularBinning()
        x.Init(buf, n + offset)
        return x

    # SparseRegularBinning
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SparseRegularBinning
    def Bins(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # SparseRegularBinning
    def BinsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # SparseRegularBinning
    def BinsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparseRegularBinning
    def BinWidth(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # SparseRegularBinning
    def Origin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float64Flags, o + self._tab.Pos)
        return 0.0

    # SparseRegularBinning
    def Overflow(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .RealOverflow import RealOverflow
            obj = RealOverflow()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SparseRegularBinning
    def LowInclusive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return True

    # SparseRegularBinning
    def HighInclusive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # SparseRegularBinning
    def Minbin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return -9223372036854775808

    # SparseRegularBinning
    def Maxbin(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 9223372036854775807

def SparseRegularBinningStart(builder): builder.StartObject(8)
def SparseRegularBinningAddBins(builder, bins): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(bins), 0)
def SparseRegularBinningStartBinsVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def SparseRegularBinningAddBinWidth(builder, binWidth): builder.PrependFloat64Slot(1, binWidth, 0.0)
def SparseRegularBinningAddOrigin(builder, origin): builder.PrependFloat64Slot(2, origin, 0.0)
def SparseRegularBinningAddOverflow(builder, overflow): builder.PrependStructSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(overflow), 0)
def SparseRegularBinningAddLowInclusive(builder, lowInclusive): builder.PrependBoolSlot(4, lowInclusive, 1)
def SparseRegularBinningAddHighInclusive(builder, highInclusive): builder.PrependBoolSlot(5, highInclusive, 0)
def SparseRegularBinningAddMinbin(builder, minbin): builder.PrependInt64Slot(6, minbin, -9223372036854775808)
def SparseRegularBinningAddMaxbin(builder, maxbin): builder.PrependInt64Slot(7, maxbin, 9223372036854775807)
def SparseRegularBinningEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class ExternalSource(object):
    external_memory = 0
    external_samefile = 1
    external_file = 2
    external_url = 3




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

# ///////////////////////////////////////////////// buffers
class DType(object):
    dtype_none = 0
    dtype_bool = 1
    dtype_int8 = 2
    dtype_uint8 = 3
    dtype_int16 = 4
    dtype_uint16 = 5
    dtype_int32 = 6
    dtype_uint32 = 7
    dtype_int64 = 8
    dtype_uint64 = 9
    dtype_float32 = 10
    dtype_float64 = 11




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class ColumnChunk(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsColumnChunk(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ColumnChunk()
        x.Init(buf, n + offset)
        return x

    # ColumnChunk
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ColumnChunk
    def Pages(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Page import Page
            obj = Page()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ColumnChunk
    def PagesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ColumnChunk
    def PageOffsets(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # ColumnChunk
    def PageOffsetsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # ColumnChunk
    def PageOffsetsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ColumnChunk
    def PageMin(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Extremes import Extremes
            obj = Extremes()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ColumnChunk
    def PageMinLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ColumnChunk
    def PageMax(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Extremes import Extremes
            obj = Extremes()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ColumnChunk
    def PageMaxLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def ColumnChunkStart(builder): builder.StartObject(4)
def ColumnChunkAddPages(builder, pages): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(pages), 0)
def ColumnChunkStartPagesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ColumnChunkAddPageOffsets(builder, pageOffsets): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(pageOffsets), 0)
def ColumnChunkStartPageOffsetsVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def ColumnChunkAddPageMin(builder, pageMin): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(pageMin), 0)
def ColumnChunkStartPageMinVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ColumnChunkAddPageMax(builder, pageMax): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(pageMax), 0)
def ColumnChunkStartPageMaxVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ColumnChunkEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import sys
import numbers
import collections
try:
    from collections.abc import Iterable
    from collections.abc import Sequence
    from collections.abc import Mapping
except ImportError:
    from collections import Iterable
    from collections import Sequence
    from collections import Mapping

import numpy

def setparent(parent, value):
    import aghast.interface

    if isinstance(value, aghast.interface.Ghast):
        if getattr(value, "_parent", parent) is not parent:
            raise ValueError("already attached to another hierarchy: {0}".format(repr(value)))
        else:
            value._parent = parent

    elif ((sys.version_info[0] >= 3 and isinstance(value, (str, bytes))) or (sys.version_info[0] < 3 and isinstance(value, basestring))):
        pass

    elif isinstance(value, Vector):
        for x in value:
            setparent(parent, x)

    elif isinstance(value, Lookup):
        for x in value.values():
            setparent(parent, x)

def _checkitem(check):
    if check.type is str:
        return CheckString(check.classname, check.paramname, required=check.required)
    elif check.type is float:
        return CheckNumber(check.classname, check.paramname, required=check.required)
    elif check.type is int:
        return CheckInteger(check.classname, check.paramname, required=check.required)
    elif isinstance(check.type, list):
        return CheckEnum(check.classname, check.paramname, required=check.required, choices=check.type)
    else:
        return CheckClass(check.classname, check.paramname, required=check.required, type=check.type)

class Vector(Sequence):
    def __init__(self, data):
        if data is None:
            self._data = ()
        else:
            self._data = tuple(data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, where):
        return self._data[where]

    def _detached(self, top):
        import aghast.interface
        return Vector([x if not isinstance(x, aghast.interface.Ghast) else x._detached(False) for x in self])

    def __repr__(self):
        tmp = [repr(x) for x in self]
        if sum(len(x) for x in tmp) < 100:
            return "[" + ", ".join(tmp) + "]"
        elif len(tmp) < 100:
            return "[" + ",\n ".join(tmp) + "]"
        else:
            return "[" + ",\n ".join(tmp[:50]) + ",\n ...\n " + ",\n ".join(tmp[:50]) + "]"

    def __eq__(self, other):
        if not isinstance(other, (Vector, Iterable)):
            return False
        if len(self) != len(other):
            return False
        for i in range(len(self)):
            if self[i] != other[i]:
                return False
        else:
            return True

    def __ne__(self, other):
        return not self.__eq__(other)

class FBVector(Vector):
    def __init__(self, length, get, check, parent):
        self._got = [None] * length
        self._get = get
        assert isinstance(check, CheckVector), repr(type(check))
        if not check.minlen <= length <= check.maxlen:
            raise TypeError("{0}.{1} length must be between {2} and {3} (inclusive)".format(check.classname, check.paramname, check.minlen, check.maxlen))
        self._check = _checkitem(check)
        self._parent = parent

    def __len__(self):
        return len(self._got)

    def __getitem__(self, where):
        if self._got[where] is None:
            self._got[where] = self._check.fromflatbuffers(self._get(where))
            setparent(self._parent, self._got[where])
        return self._got[where]

class Lookup(Mapping):
    def __init__(self, data):
        import aghast.interface
        if data is None:
            self._data = collections.OrderedDict()
        else:
            self._data = collections.OrderedDict(data)
        for n, x in self._data.items():
            if isinstance(x, aghast.interface.Ghast) and not hasattr(x, "_parent"):
                x._identifier = n

    def __len__(self):
        return len(self._data)

    def __getitem__(self, where):
        return self._data[where]

    def __iter__(self):
        return iter(self._data)

    def _detached(self, top):
        import aghast.interface
        return Lookup({n: x if not isinstance(x, aghast.interface.Ghast) else x._detached(False) for n, x in self.items()})

    def __repr__(self):
        tmp = [repr(n) + ": " + repr(x) for n, x in self.items()]
        if sum(len(x) for x in tmp) < 100:
            return "{" + ", ".join(tmp) + "}"
        elif len(tmp) < 100:
            return "{" + ",\n ".join(tmp) + "}"
        else:
            return "{" + ",\n ".join(tmp[:50]) + ",\n ...\n " + ",\n ".join(tmp[:50]) + "}"

    def __eq__(self, other):
        if not isinstance(other, (Lookup, Mapping)):
            return False
        if set(self) != set(other):
            return False
        for n in self:
            if self[n] != other[n]:
                return False
        else:
            return True

    def __ne__(self, other):
        return not self.__eq__(other)

class FBLookup(Lookup):
    def __init__(self, length, lookup, get, check, parent):
        self._lookup = {lookup(i).decode("utf-8"): i for i in range(length)}
        self._got = {}
        self._get = get
        assert isinstance(check, CheckLookup), repr(type(check))
        if not check.minlen <= length <= check.maxlen:
            raise TypeError("{0}.{1} length must be between {2} and {3} (inclusive)".format(check.classname, check.paramname, check.minlen, check.maxlen))
        self._check = _checkitem(check)
        self._parent = parent

    def __len__(self):
        return len(self._lookup)

    def __iter__(self):
        return iter(self._lookup)

    def __getitem__(self, where):
        import aghast.interface

        item = self._got.get(where, None)
        if item is None:
            item = self._check.fromflatbuffers(self._get(self._lookup[where]))
            self._got[where] = item
            if isinstance(item, aghast.interface.Ghast):
                item._identifier = where
            setparent(self._parent, item)
        return item

class Check(object):
    def __init__(self, classname, paramname, required):
        self.classname = classname
        self.paramname = paramname
        self.required = required

    def __repr__(self):
        return "<{0} {1}.{2} at 0x{3:012x}>".format(type(self).__name__, self.classname, self.paramname, id(self))

    def __call__(self, obj):
        if obj is None and self.required:
            raise TypeError("{0}.{1} is required, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
        else:
            return obj

    def fromflatbuffers(self, obj):
        return obj

class CheckBool(Check):
    def __call__(self, obj):
        super(CheckBool, self).__call__(obj)
        if obj is None:
            return obj
        elif not isinstance(obj, (bool, numpy.bool_, numpy.bool)):
            raise TypeError("{0}.{1} must be boolean, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
        return bool(obj)

class CheckString(Check):
    def __call__(self, obj):
        super(CheckString, self).__call__(obj)
        if obj is None:
            return obj
        elif not ((sys.version_info[0] >= 3 and isinstance(obj, str)) or (sys.version_info[0] < 3 and isinstance(obj, basestring))):
            raise TypeError("{0}.{1} must be a string, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
        else:
            return obj

    def fromflatbuffers(self, obj):
        if obj is None:
            return obj
        else:
            return obj.decode("utf-8")

class CheckNumber(Check):
    def __init__(self, classname, paramname, required, min=float("-inf"), max=float("inf"), min_inclusive=True, max_inclusive=True):
        super(CheckNumber, self).__init__(classname, paramname, required)
        self.min = min
        self.max = max
        self.min_inclusive = min_inclusive
        self.max_inclusive = max_inclusive

    def __call__(self, obj):
        super(CheckNumber, self).__call__(obj)
        if obj is None:
            return obj
        elif not isinstance(obj, (numbers.Real, numpy.floating, numpy.integer)) or numpy.isnan(obj):
            raise TypeError("{0}.{1} must be a number, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
        elif self.min_inclusive and not self.min <= obj:
            raise TypeError("{0}.{1} must not be below {2} (inclusive), cannot pass {3} (type {4})".format(self.classname, self.paramname, self.min, repr(obj), type(obj)))
        elif not self.min_inclusive and not self.min < obj:
            raise TypeError("{0}.{1} must not be below {2} (exclusive), cannot pass {3} (type {4})".format(self.classname, self.paramname, self.min, repr(obj), type(obj)))
        elif self.max_inclusive and not obj <= self.max:
            raise TypeError("{0}.{1} must not be above {2} (inclusive), cannot pass {3} (type {4})".format(self.classname, self.paramname, self.max, repr(obj), type(obj)))
        elif not self.max_inclusive and not obj < self.max:
            raise TypeError("{0}.{1} must not be above {2} (exclusive), cannot pass {3} (type {4})".format(self.classname, self.paramname, self.max, repr(obj), type(obj)))
        else:
            return float(obj)

class CheckInteger(Check):
    def __init__(self, classname, paramname, required, min=float("-inf"), max=float("inf")):
        super(CheckInteger, self).__init__(classname, paramname, required)
        self.min = min
        self.max = max

    def __call__(self, obj):
        super(CheckInteger, self).__call__(obj)
        if obj is None:
            return obj
        elif not isinstance(obj, (numbers.Integral, numpy.integer)):
            raise TypeError("{0}.{1} must be an integer, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
        elif not self.min <= obj:
            raise TypeError("{0}.{1} must not be below {2} (inclusive), cannot pass {3} (type {4})".format(self.classname, self.paramname, self.min, repr(obj), type(obj)))
        elif not obj <= self.max:
            raise TypeError("{0}.{1} must not be above {2} (inclusive), cannot pass {3} (type {4})".format(self.classname, self.paramname, self.max, repr(obj), type(obj)))
        else:
            return int(obj)

class CheckEnum(Check):
    def __init__(self, classname, paramname, required, choices, intlookup=None):
        super(CheckEnum, self).__init__(classname, paramname, required)
        self.choices = choices
        self.intlookup = intlookup

    def __call__(self, obj):
        super(CheckEnum, self).__call__(obj)
        if obj is None:
            return obj
        elif obj not in self.choices:
            raise TypeError("{0}.{1} must be one of {2}, cannot pass {3} (type {4})".format(self.classname, self.paramname, self.choices, repr(obj), type(obj)))
        else:
            return self.choices[self.choices.index(obj)]

    def fromflatbuffers(self, obj):
        if obj is None:
            return obj
        elif self.intlookup is None:
            return self.choices[obj]
        else:
            return self.intlookup[obj]

class CheckClass(Check):
    def __init__(self, classname, paramname, required, type):
        super(CheckClass, self).__init__(classname, paramname, required)
        self.type = type

    def __call__(self, obj):
        super(CheckClass, self).__call__(obj)
        if obj is None:
            return obj
        elif not isinstance(obj, self.type):
            raise TypeError("{0}.{1} must be a {2} object, cannot pass {3} (type {4})".format(self.classname, self.paramname, self.type, repr(obj), type(obj)))
        return obj

    def fromflatbuffers(self, obj):
        if obj is None:
            return obj
        else:
            return self.type._fromflatbuffers(obj)

class CheckKey(Check):
    def __init__(self, classname, paramname, required, type):
        super(CheckKey, self).__init__(classname, paramname, required)
        self.type = type

    def __call__(self, obj):
        super(CheckKey, self).__call__(obj)
        if obj is None:
            return obj
        elif self.type is str:
            if not ((sys.version_info[0] >= 3 and isinstance(obj, str)) or (sys.version_info[0] < 3 and isinstance(obj, basestring))):
                raise TypeError("{0}.{1} must be a string, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
            return obj
        elif self.type is float:
            if not isinstance(obj, (numbers.Real, numpy.floating, numpy.integer)):
                raise TypeError("{0}.{1} must be a number, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
            return float(obj)
        elif self.type is int:
            if not isinstance(obj, (numbers.Integral, numpy.integer)):
                raise TypeError("{0}.{1} must be an integer, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
            return int(obj)
        else:
            if not isinstance(obj, self.type):
                raise TypeError("{0}.{1} must be a {2} object, cannot pass {3} (type {4})".format(self.classname, self.paramname, self.type, repr(obj), type(obj)))
            return obj

    def fromflatbuffers(self, obj):
        if obj is None or self.type is float or self.type is int:
            return obj
        elif self.type is str:
            return obj.decode("utf-8")
        else:
            return self.type._fromflatbuffers(obj)

class CheckVector(Check):
    def __init__(self, classname, paramname, required, type, minlen=0, maxlen=float("inf")):
        super(CheckVector, self).__init__(classname, paramname, required)
        self.type = type
        self.minlen = minlen
        self.maxlen = maxlen

    def __call__(self, obj):
        super(CheckVector, self).__call__(obj)
        if obj is None:
            return Vector(obj)
        elif not isinstance(obj, Iterable):
            raise TypeError("{0}.{1} must be iterable, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))

        if ((sys.version_info[0] >= 3 and isinstance(obj, (str, bytes))) or (sys.version_info[0] < 3 and isinstance(obj, basestring))):
            raise TypeError("{0}.{1} must be iterable but not a string, cannot pass {2}".format(self.classname, self.paramname, repr(obj)))

        if not self.minlen <= len(obj) <= self.maxlen:
            raise TypeError("{0}.{1} length must be between {2} and {3} (inclusive), cannot pass {4} (type ({5}))".format(self.classname, self.paramname, self.minlen, self.maxlen, repr(obj), type(obj)))

        if self.type is str:
            for x in obj:
                if not ((sys.version_info[0] >= 3 and isinstance(x, str)) or (sys.version_info[0] < 3 and isinstance(x, basestring))):
                    raise TypeError("{0}.{1} elements must be strings, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(x), type(x)))
            return Vector(obj)
        elif self.type is int:
            return numpy.array(obj, dtype="<i8")
        elif self.type is float:
            return numpy.array(obj, dtype="<f8")
        elif isinstance(self.type, list):
            for x in obj:
                if not x in self.type:
                    raise TypeError("{0}.{1} elements must be one of {2}, cannot pass {3} (type {4})".format(self.classname, self.paramname, self.type, repr(x), type(x)))
            return Vector(self.type[self.type.index(x)] for x in obj)
        else:
            for x in obj:
                if not isinstance(x, self.type):
                    raise TypeError("{0}.{1} elements must be {2} objects, cannot pass {3} (type {4})".format(self.classname, self.paramname, self.type, repr(x), type(x)))
            return Vector(obj)

class CheckLookup(Check):
    def __init__(self, classname, paramname, required, type, minlen=0, maxlen=float("inf")):
        super(CheckLookup, self).__init__(classname, paramname, required)
        self.type = type
        self.minlen = minlen
        self.maxlen = maxlen

    def __call__(self, obj):
        super(CheckLookup, self).__call__(obj)
        if obj is None:
            return Lookup(obj)
        elif not isinstance(obj, (dict, Mapping)):
            raise TypeError("{0}.{1} must be a dict or Mapping, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))

        if not self.minlen <= len(obj) <= self.maxlen:
            raise TypeError("{0}.{1} length must be between {2} and {3} (inclusive), cannot pass {4} (type ({5}))".format(self.classname, self.paramname, self.minlen, self.maxlen, repr(obj), type(obj)))

        if self.type is str:
            for x in obj.values():
                if not ((sys.version_info[0] >= 3 and isinstance(x, str)) or (sys.version_info[0] < 3 and isinstance(x, basestring))):
                    raise TypeError("{0}.{1} elements must be strings, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(x), type(x)))
            return Lookup(obj)
        elif self.type is float:
            for x in obj.values():
                if not isinstance(x, (numbers.Real, numpy.integer, numpy.floating)):
                    raise TypeError("{0}.{1} elements must be numbers, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(x), type(x)))
            return Lookup(float(x) for x in obj)
        elif isinstance(self.type, list):
            for x in obj.values():
                if not x in self.type:
                    raise TypeError("{0}.{1} elements must be one of {2}, cannot pass {3} (type {4})".format(self.classname, self.paramname, self.type, repr(x), type(x)))
            return Lookup(self.type[self.type.index(x)] for x in obj)
        else:
            for x in obj.values():
                if not isinstance(x, self.type):
                    raise TypeError("{0}.{1} elements must be {2} objects, cannot pass {3} (type {4})".format(self.classname, self.paramname, self.type, repr(x), type(x)))
            return Lookup(obj)

class CheckBuffer(Check):
    def __call__(self, obj):
        super(CheckBuffer, self).__call__(obj)
        if obj is None:
            return obj
        if isinstance(obj, numpy.ndarray) and not obj.flags.c_contiguous:
            obj = obj.copy(order="C")
        try:
            return numpy.frombuffer(obj, dtype=numpy.uint8)
        except AttributeError:
            raise TypeError("{0}.{1} must be a buffer, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))

class CheckSlice(Check):
    def __call__(self, obj):
        super(CheckSlice, self).__call__(obj)
        if obj is None:
            return obj
        elif not isinstance(obj, slice):
            raise TypeError("{0}.{1} must be a slice, cannot pass {2} (type {3})".format(self.classname, self.paramname, repr(obj), type(obj)))
        return out

    def fromflatbuffers(self, obj):
        if obj is None:
            return obj
        else:
            return slice(obj.Start() if obj.HasStart() else None, obj.Stop() if obj.HasStop() else None, obj.Step() if obj.HasStep() else None)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class ParameterizedFunction(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsParameterizedFunction(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ParameterizedFunction()
        x.Init(buf, n + offset)
        return x

    # ParameterizedFunction
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ParameterizedFunction
    def Expression(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ParameterizedFunction
    def Parameters(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Parameter import Parameter
            obj = Parameter()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ParameterizedFunction
    def ParametersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ParameterizedFunction
    def Paramaxis(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ParameterizedFunction
    def ParamaxisAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # ParameterizedFunction
    def ParamaxisLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ParameterizedFunction
    def ParameterCovariances(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Covariance import Covariance
            obj = Covariance()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ParameterizedFunction
    def ParameterCovariancesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def ParameterizedFunctionStart(builder): builder.StartObject(4)
def ParameterizedFunctionAddExpression(builder, expression): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(expression), 0)
def ParameterizedFunctionAddParameters(builder, parameters): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(parameters), 0)
def ParameterizedFunctionStartParametersVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ParameterizedFunctionAddParamaxis(builder, paramaxis): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(paramaxis), 0)
def ParameterizedFunctionStartParamaxisVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ParameterizedFunctionAddParameterCovariances(builder, parameterCovariances): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(parameterCovariances), 0)
def ParameterizedFunctionStartParameterCovariancesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ParameterizedFunctionEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import math
import numbers
try:
    from inspect import signature
except ImportError:
    try:
        from funcsigs import signature
    except ImportError:
        raise ImportError("Install funcsigs package with:\n    pip install funcsigs\nor\n    conda install funcsigs\n(or just use Python >= 3.3).")

import numpy

import aghast
import aghast.checktype
import aghast.interface

prologue = u"""= Aghast Specification

== Introduction

Aghast encodes aggregated, histogram-like statistics in a way that can be shared among many libraries that fill, fit, or plot such things. It represents reduced data, such as the output of a histogram-filling (map) and combining (reduce) process, or the output of an SQL-like group-by procedure.

Aghast is not intended to be used directly by data analysts. It is horrifically verbose. Instead, it is a general ontology of aggregated data types with conversion routines to translate among libraries. With the appropriate translations behind the scenes, data analysts can reduce large datasets with one library, model/fit the aggregation with another, and plot it with a third.

This role is inspired by two projects: https://arrow.apache.org[Apache Arrow], which acts as an efficient memory representation for nested, columnar data, and http://dmg.org/pmml/v4-3/GeneralStructure.html[PMML], which is an ontology of data mining models in XML. Like Arrow, aghast is an in-memory format to pass data in a random-access and zero-copy way (though an in-memory format can also be saved to disk). Like PMML, aghast expresses domain-specific statistical concepts, such as histograms, moments, quantiles, covariances, fit functions, and ntuples, rather than generic data types.

Aghast is specified using https://google.github.io/flatbuffers[Google Flatbuffers]. Flatbuffers allows for lazy, random-access interpretation: any nested property can be accessed without reading or deserializing the others. Thus, a "`ghast`" (a sharable object) could contain thousands of histograms and accessing the bin contents of a particular one would not only leave the other histograms untouched, it would leave the histogram's own metadata untouched. Flatbuffers minimizes its serialized size (because it is intended as a network protocol) and its serialization/deserialization time (because it is intended for computer games).

Aghast is designed to scale. Modern particle physics analyses deal in thousands of histograms generated by viewing the same quantities with many different "`cuts`" (selection filters) and variations (to test senstivity to systematic effects). This usually results in duplicated histogram axis descriptions: every histogram in a group of a hundred must be binned the same way. Duplication inflates serialized size, but it also introduces the possibility that some binnings might disagree (due to a copy-paste error) and the need for extra validity checks. Aghast is designed around the idea of "`superhistograms,`" representing groups of histograms that have identical binnings as a binning in a larger histogram. Thus, there are axis types for <<PredicateBinning>> (defined by if-then predicates) and <<VariationBinning>> (defined by systematic variations), and the data filling these bins are contiguous across all histograms in the group.

What this specification does not define are any methods for filling, fitting, or plotting. From a data analyst's perspective, this is everything one does with histograms or other statistical objects. This is because aghast is firmly behind-the-scenes, a helper to other fine analysis tools, such as https://root.cern[ROOT], https://github.com/boostorg/histogram[Boost.Histogram], https://physt.readthedocs.io/en/latest/index.html[Physt], and https://pandas.pydata.org[Pandas].

== Data Types

Aghast is specified as a Flatbuffers format in link:flatbuffers/aghast.fbs[]. Flatbuffers provides a standard suite of types that can be translated into https://google.github.io/flatbuffers/flatbuffers_support.html[many languages]. However, the code the Flatbuffers code generator produces is too low-level even for applications to use as a backend, so we describe the interfaces and types of wrapper classes here.

These class descriptions are sufficiently constrained to fit into any static type system, and while they include heterogeneous lists (lists of an enumerated union type), they can be decomposed into homogeneous lists with an additional level of nesting. In fact, the Flatbuffers specification doesn't allow heterogeneous lists, so it provides an example of this decomposition. (For example, the heterogeneous *objects* list in a <<Collection>>, which can contain <<Histogram>>, <<ParameterizedFunction>>, <<BinnedEvaluatedFunction>>, and <<Ntuple>>, is encoded in Flatbuffers as a homogeneous list of `Object`, which contains a _single_ union of `ObjectData`.)

Basic types, like booleans, integers, floating point numbers, and strings, are passed through without modification (though strings are explicitly encoded as `utf-8`; Flatbuffers strings are not encoding-aware). Integers and floating point numbers may have a constrained range, such as [0, \u221e) for non-negative numbers excluding \u221e or (0, 2\u03c0] for positive values less than or equal to 2\u03c0. (A square bracket includes the endpoint; a round bracket excludes it.) Empty strings and missing strings (null) are distinct.

Lists may contain basic types or class instances, and there is no distinction between empty lists and missing lists (an artifact of Flatbuffers).

In some cases, we want a mapping type, such as str \u2192 X, so that objects are retrievable by name, rather than index. Flatbuffers does not have such a type, so we build it by decomposing the high-level mapping into a low-level pair of lists with equal length. (For example, the *objects* mapping in a <<Collection>> is encoded in Flatbuffers as a list `objects` and a list `lookup`.)

Class objects may be missing (null) if they are not required. Required properties are a Flatbuffers feature: it doesn't generate code that would allow the serialized object to be missing. The class schemas can evolve to include more properties (with full forward and backward compatibility), but properties cannot be removed and required properties cannot become non-required.

Any properties that are not required have a default value (usually null).

In addition to the type constraints, which are tighter in the wrapper classes than they are in the Flatbuffers serialization, we list invariants (conditions that must be true) that depend on multiple properties or multiple class objects. To avoid unnecessary Flatbuffers deserialization, these are not automatically checked, but a check can be invoked and any behavior when those invariants are not satisfied is undefined.

A sharable ghast may have one of the following types: <<Collection>>, <<Histogram>>, <<ParameterizedFunction>>, <<BinnedEvaluatedFunction>>, and <<Ntuple>>."""

epilogue = u""

classes = [
    aghast.Collection,
    aghast.Histogram,
    aghast.Axis,
    aghast.IntegerBinning,
    aghast.RegularBinning,
    aghast.RealInterval,
    aghast.RealOverflow,
    aghast.HexagonalBinning,
    aghast.EdgesBinning,
    aghast.IrregularBinning,
    aghast.CategoryBinning,
    aghast.SparseRegularBinning,
    aghast.FractionBinning,
    aghast.PredicateBinning,
    aghast.VariationBinning,
    aghast.Variation,
    aghast.Assignment,
    aghast.UnweightedCounts,
    aghast.WeightedCounts,
    aghast.InterpretedInlineBuffer,
    aghast.InterpretedInlineInt64Buffer,
    aghast.InterpretedInlineFloat64Buffer,
    aghast.InterpretedExternalBuffer,
    aghast.Profile,
    aghast.Statistics,
    aghast.Moments,
    aghast.Quantiles,
    aghast.Modes,
    aghast.Extremes,
    aghast.StatisticFilter,
    aghast.Covariance,
    aghast.ParameterizedFunction,
    aghast.Parameter,
    aghast.EvaluatedFunction,
    aghast.BinnedEvaluatedFunction,
    aghast.Ntuple,
    aghast.Column,
    aghast.NtupleInstance,
    aghast.Chunk,
    aghast.ColumnChunk,
    aghast.Page,
    aghast.RawInlineBuffer,
    aghast.RawExternalBuffer,
    aghast.Metadata,
    aghast.Decoration,
    ]

unions = {
    aghast.interface.RawBuffer: [aghast.RawInlineBuffer, aghast.RawExternalBuffer],
    aghast.interface.InterpretedBuffer: [aghast.InterpretedInlineBuffer, aghast.InterpretedInlineInt64Buffer, aghast.InterpretedInlineFloat64Buffer, aghast.InterpretedExternalBuffer],
    aghast.interface.Binning: [aghast.IntegerBinning, aghast.RegularBinning, aghast.HexagonalBinning, aghast.EdgesBinning, aghast.IrregularBinning, aghast.CategoryBinning, aghast.SparseRegularBinning, aghast.FractionBinning, aghast.PredicateBinning, aghast.VariationBinning],
    aghast.interface.Counts: [aghast.UnweightedCounts, aghast.WeightedCounts],
    aghast.interface.Function: [aghast.ParameterizedFunction, aghast.EvaluatedFunction],
    aghast.interface.FunctionObject: [aghast.ParameterizedFunction, aghast.BinnedEvaluatedFunction],
    aghast.interface.Object: [aghast.Histogram, aghast.Ntuple, aghast.ParameterizedFunction, aghast.BinnedEvaluatedFunction, aghast.Collection],
    }

def num(x):
    if not isinstance(x, (bool, numpy.bool, numpy.bool)) and isinstance(x, (numbers.Real, numpy.integer, numpy.floating)):
        if x == float("-inf"):
            return u"\u2012\u221e"
        elif x == float("inf"):
            return u"\u221e"
        elif x == -0.5*math.pi:
            return u"\u2012\u03c0/2"
        elif x == 0.5*math.pi:
            return u"\u03c0/2"
        elif x == aghast.interface.MININT64:
            return u"\u20122\u2076\u00b3"  # -2**63
        elif x == aghast.interface.MAXINT64:
            return u"2\u2076\u00b3 \u2012 1"  # 2**63 - 1
        elif x == 0.5:
            return "1/2"
        elif x < 0:
            return u"\u2012" + repr(abs(x))
        else:
            return repr(abs(x))
    elif x is None:
        return "null"
    elif x is True:
        return "true"
    elif x is False:
        return "false"
    elif x == []:
        return "null/empty"
    else:
        return "`+{0}+`".format(repr(x))

def formatted(cls, end="\n"):
    out = [end + end + "== {0}{1}".format(cls.__name__, end), "*" + cls.description.strip() + "*"]

    out.append(end + "[%hardbreaks]")
    for name, param in signature(cls.__init__).parameters.items():
        if name != "self":
            check = cls._params[name]
            hasdefault = param.default is not param.empty

            islist = False
            if isinstance(check, aghast.checktype.CheckBool):
                typestring = "bool"
            elif isinstance(check, aghast.checktype.CheckString):
                typestring = "str"
            elif isinstance(check, aghast.checktype.CheckNumber):
                typestring = "float in {0}{1}, {2}{3}".format(
                    "[" if check.min_inclusive else "(",
                    num(check.min),
                    num(check.max),
                    "]" if check.max_inclusive else ")")
            elif isinstance(check, aghast.checktype.CheckInteger):
                typestring = "int in {0}{1}, {2}{3}".format(
                    "(" if check.min == float("-inf") else "[",
                    num(check.min),
                    num(check.max),
                    ")" if check.max == float("inf") else "]")
            elif isinstance(check, aghast.checktype.CheckEnum):
                typestring = "one of {" + ", ".join("`+" + str(x) + "+`" for x in check.choices) + "}"
            elif isinstance(check, aghast.checktype.CheckClass):
                if check.type in unions:
                    typestring = " or ".join("<<{0}>>".format(x.__name__) for x in unions[check.type])
                else:
                    typestring = "<<{0}>>".format(check.type.__name__)
            elif isinstance(check, aghast.checktype.CheckKey) and check.type is str:
                typestring = "unique str"
            elif isinstance(check, (aghast.checktype.CheckVector, aghast.checktype.CheckLookup)):
                islist = True
                if check.type is str:
                    subtype = "str"
                elif check.type is int:
                    subtype = "int"
                elif check.type is float:
                    subtype = "float"
                elif isinstance(check.type, list):
                    subtype = "{" + ", ".join("`+" + str(x) + "+`" for x in check.type) + "}"
                else:
                    if check.type in unions:
                        subtype = " or ".join("<<{0}>>".format(x.__name__) for x in unions[check.type])
                    else:
                        subtype = "<<{0}>>".format(check.type.__name__)
                if check.minlen != 0 or check.maxlen != float("inf"):
                    withlength = " with length in [{0}, {1}{2}".format(
                        num(check.minlen),
                        num(check.maxlen),
                        ")" if check.maxlen == float("inf") else "]")
                else:
                    withlength = ""
                if isinstance(check, aghast.checktype.CheckVector):
                    typestring = "list of {0}{1}".format(subtype, withlength)
                else:
                    typestring = u"str \u2192 {0}{1}".format(subtype, withlength)
            elif isinstance(check, aghast.checktype.CheckBuffer):
                typestring = "buffer"
            elif isinstance(check, aghast.checktype.CheckSlice):
                typestring = "slice (start:stop:step)"
            else:
                raise AssertionError(type(check))

            linebreak = " +" + end if len(name) + len(typestring) > 75 else " "
            if check.required:
                defaultstring = "(required)"
            elif hasdefault:
                defaultstring = "(default: {0})".format(num([] if islist and param.default is None else param.default))
            else:
                raise AssertionError

            out.append(u"\u2022{nbsp}" + " *{0}*: {1}{2}{3}".format(name, typestring, linebreak, defaultstring))

    if len(cls.validity_rules) != 0:
        out.append((" +" + end).join(cls.validity_rules))

    if cls.long_description is not None:
        out.append(end + "*Details:*" + end)
        out.append(cls.long_description.strip())

    return end.join(out)

if __name__ == "__main__":
    with open("../specification.adoc", "w") as file:
        file.write(prologue)

        for cls in classes:
            file.write(formatted(cls))

        file.write(epilogue)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class WeightedCounts(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsWeightedCounts(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = WeightedCounts()
        x.Init(buf, n + offset)
        return x

    # WeightedCounts
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # WeightedCounts
    def SumwType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # WeightedCounts
    def Sumw(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # WeightedCounts
    def Sumw2Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # WeightedCounts
    def Sumw2(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # WeightedCounts
    def Unweighted(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .UnweightedCounts import UnweightedCounts
            obj = UnweightedCounts()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def WeightedCountsStart(builder): builder.StartObject(5)
def WeightedCountsAddSumwType(builder, sumwType): builder.PrependUint8Slot(0, sumwType, 0)
def WeightedCountsAddSumw(builder, sumw): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(sumw), 0)
def WeightedCountsAddSumw2Type(builder, sumw2Type): builder.PrependUint8Slot(2, sumw2Type, 0)
def WeightedCountsAddSumw2(builder, sumw2): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(sumw2), 0)
def WeightedCountsAddUnweighted(builder, unweighted): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(unweighted), 0)
def WeightedCountsEnd(builder): return builder.EndObject()
